﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using OverStore.Exceptions;
using OverStore.Runtime.ObjectRepository;

namespace OverStore.Runtime.PersistenceServiceProvider
{
    /// <summary>
    /// Base class allows to virtual default implementation of <see cref="IPersistenceServiceProvider"/>.
    /// Default implementation of each method calls corresponding method of BaseProvider.
    /// </summary>
    public abstract class CustomPersistenceServiceProvider : IPersistenceServiceProvider
    {
        private readonly IPersistenceServiceProvider _BaseProvider;

        /// <summary>
        /// Initializes a new instance of the <see cref="CustomPersistenceServiceProvider"/> class.
        /// </summary>
        /// <param name="baseProvider">The base provider.</param>
        public CustomPersistenceServiceProvider(IPersistenceServiceProvider baseProvider)
        {
            if (Object.ReferenceEquals(baseProvider, null))
                throw new ArgumentNullException("baseProvider");
            _BaseProvider = baseProvider;
        }

        /// <summary>
        /// Gets the base provider which method should be usually called.
        /// </summary>
        protected IPersistenceServiceProvider BaseProvider
        {
            get
            {
                return _BaseProvider;
            }
        }

        /// <summary>
        /// 	<para> Updates data of specified instance using data extracted from data storage, reload instance aggregates and
        /// mark instance as unchanged, so further saving will not cause data storage updating (except if object was changed).
        /// </para>
        /// 	<para> If specified <paramref name="instance"/> is not registered in repository, method also registers it.
        /// If deffer loading is supported and enabled, the object will only be registered for
        /// loading, and loading itself will be performed by request.</para>
        /// </summary>
        /// <typeparam name="TInstance"></typeparam>
        /// <param name="instance">The instance to reload.</param>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same  already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error is occured during reload object data.</exception>
        public virtual bool Reload<TInstance>(TInstance instance)
            where TInstance : class
        {
            return BaseProvider.Reload<TInstance>(instance);
        }

        /// <summary>
        /// Clears collection and fills it up with object read from parameterless reader for instance.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <exception cref="IdentityReferenceIntegrityException">Collection contains instance of object with same key already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error is occured during reload object data.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void ReloadCollection<TInstance>(IEnumerable<TInstance> collection)
            where TInstance : class
        {
            BaseProvider.ReloadCollection<TInstance>(collection);
        }

        /// <summary>
        /// Clears collection and fills it up with object read from parameterless reader for instance.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <typeparam name="TParameter">The type of the parameter.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <param name="parameter">The parameter used for get data for loading.</param>
        /// <exception cref="IdentityReferenceIntegrityException">Collection contains instance of object with same key already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error is occured during reload object data.</exception>
        /// <exception cref="ParameterTypeNotSupportedException">Parameters of specified type not supported.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void ReloadCollection<TInstance, TParameter>(IEnumerable<TInstance> collection, TParameter parameter)
            where TInstance : class
            where TParameter : class
        {
            BaseProvider.ReloadCollection<TInstance, TParameter>(collection, parameter);
        }

        /// <summary>
        /// 	<para>
        /// Saves changes made in <paramref name="instance"/> in data storage and invokes saving instance's aggregates.
        /// If instance was previously marked as deleted, method removes it from data storage.
        /// </para>
        /// 	<para>
        /// If instance supports deffer loading and is not loaded yet, method does not perform any
        /// actions, because changes can not be made without instance loading.
        /// </para>
        /// 	<para>
        /// Method does not accept instance changes, but types implements current method should consider possibility
        /// of calling this method few times with same instance, in case of cycle references, for example.
        /// Repeated calls should not cause repeated saving, so method should take steps to avoid repeated saving.
        /// For example, instance may internally maintain list of saved instances, checks this list on saving and
        /// clear this list in see cref="IRepository{TInstance}.AcceptChanges" method.
        /// </para>
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="instance">The instance to save changes.</param>
        /// <exception cref="PersistingRuntimeException">Error occured while save object.</exception>
        /// <exception cref="VersionConflictException">Version conflict occured during save object.</exception>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same identifier already registered.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void Save<TInstance>(TInstance instance)
            where TInstance : class
        {
            BaseProvider.Save<TInstance>(instance);
        }

        /// <summary>
        /// 	<para>
        /// Saves changes made in each item of collection to data storage and invokes saving instance's aggregates.
        /// If instance was previously marked as deleted, method removes it from data storage.
        /// Also method saves instances were removed from collection, if collection type persister supports this feature.
        /// </para>
        /// 	<para>
        /// If collection supports deffer loading and is not loaded yet, method does not perform any
        /// actions, because changes can not be made without collection loading.
        /// </para>
        /// 	<para>
        /// Method does not accept collection changes, but types implements current method should consider possibility
        /// of calling this method few times with same instance, in case of cycle references, for example.
        /// Repeated calls should not cause repeated saving, so method should take steps to avoid repeated saving.
        /// For example, instance may internally maintain list of saved instances, checks this list on saving and
        /// clear this list in see cref="IRepository{TInstance}.AcceptChanges" method.
        /// </para>
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <exception cref="PersistingRuntimeException">Error occured while save object.</exception>
        /// <exception cref="VersionConflictException">Version conflict occured during save object.</exception>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same identifier already registered.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void SaveCollection<TInstance>(IEnumerable<TInstance> collection)
            where TInstance : class
        {
            BaseProvider.SaveCollection<TInstance>(collection);
        }

        /// <summary>
        /// Marks <paramref name="instance"/> as deleted.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="instance">The instance to mark as deleted.</param>
        /// <remarks>
        /// Method itself does not save any changes to storage. For actually delete <paramref name="instance"/>
        /// from storage, call <see cref="IRepository{TInstance}.Save"/> method.
        /// </remarks>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same identifier already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error occured while mark object as deleted.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void MarkAsDeleted<TInstance>(TInstance instance)
            where TInstance : class
        {
            BaseProvider.MarkAsDeleted<TInstance>(instance);
        }

        /// <summary>
        /// Marks every item in collection as deleted. Method does not clears collection content.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <remarks>
        /// Method itself does not save any changes to storage. For actually delete <paramref name="instance"/>
        /// from storage, call <see cref="IRepository{TInstance}.Save"/> method.
        /// </remarks>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same identifier already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error occured while mark object as deleted.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void MarkCollectionAsDeleted<TInstance>(IEnumerable<TInstance> collection)
            where TInstance : class
        {
            BaseProvider.MarkCollectionAsDeleted<TInstance>(collection);
        }

        /// <summary>
        /// Marks <paramref name="instance"/> as needed to be physically saved to storage even
        /// if <paramref name="instance"/> data are not changed.
        /// This method may be called any times untill AcceptChanges method called, but instance will be saved only once.
        /// </summary>
        /// <typeparam name="TInstance">Type of instance.</typeparam>
        /// <param name="instance">Instance to mark as dirty.</param>
        /// <remarks>
        /// Use this method in couple with optimistic locking (version checking) to organize object hierarchy into 'atom' group.
        /// Call this method on successful instance saving using other members of 'atom' object group, then save these objects.
        /// </remarks>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with the same identifier already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error occured while mark object as dirty.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void MarkAsDirty<TInstance>(TInstance instance)
            where TInstance : class
        {
            BaseProvider.MarkAsDirty<TInstance>(instance);
        }

        /// <summary>
        /// Marks every item in collection as needed to be physically saved to storage even if item data are not changed.
        /// This method may be called any times untill AcceptChanges method called, but instance will be saved only once.
        /// </summary>
        /// <typeparam name="TInstance">Type of instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <remarks>
        /// Use this method in couple with optimistic locking (version checking) to organize objects hierarchy into 'atom' group.
        /// Call this method on successful instance saving using other members of 'atom' object group, then save these objects.
        /// </remarks>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with the same identifier already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error occured while mark object as dirty.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void MarkCollectionAsDirty<TInstance>(IEnumerable<TInstance> collection)
            where TInstance : class
        {
            BaseProvider.MarkCollectionAsDirty<TInstance>(collection);
        }

        /// <summary>
        /// 	<para> Registers <paramref name="instance"/> in repository, if it is not registered yet. After registration
        /// <paramref name="instance"/> become available by its identifier. </para>
        /// 	<para> If same instance already registered in repository, method does not perform any action.</para>
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="instance">The instance to registers in repository.</param>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same identifier already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error occured during add object.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void Add<TInstance>(TInstance instance)
            where TInstance : class
        {
            BaseProvider.Add<TInstance>(instance);
        }

        /// <summary>
        /// 	<para> Registers every item of collection in session , if it is not registered yet.
        /// After registration item become available by its identifier. </para>
        /// 	<para> If same instance already registered in repository, method does not perform any action.</para>
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same identifier already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error occured during add object.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void AddCollection<TInstance>(IEnumerable<TInstance> collection)
            where TInstance : class
        {
            BaseProvider.AddCollection<TInstance>(collection);
        }

        /// <summary>
        /// Joins specified instance or collection to repository if required, applies changes made to <paramref name="instance"/> and attach aggregates of specified instance.
        /// If specified instance is collection, attach each items and deletes items deleted from collection since detaching.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="instance">The instance to attach.</param>
        /// <returns>
        /// The instance registered in repository contains data of <paramref name="instance"/>.
        /// </returns>
        /// <exception cref="PersistingRuntimeException">Error occured during Attach object.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual TInstance Attach<TInstance>(TInstance instance)
            where TInstance : class
        {
            return BaseProvider.Attach<TInstance>(instance);
        }

        /// <summary>
        /// Joins every item of collection to repository if required, applies changes made to <paramref name="instance"/> and attach aggregates of specified instance.
        /// If specified instance is collection, attach each items and deletes items deleted from collection since detaching.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <returns>
        /// The instance registered in repository contains data of <paramref name="instance"/>.
        /// </returns>
        /// <exception cref="PersistingRuntimeException">Error occured during Attach object.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual IEnumerable<TInstance> AttachCollection<TInstance>(IEnumerable<TInstance> collection)
            where TInstance : class
        {
            return BaseProvider.AttachCollection<TInstance>(collection);
        }

        /// <summary>
        /// Get identifier for object registered in session.
        /// If object is not registered in session, it will be placed in session.
        /// If object does not have identifier yet, the identifier will be assigned to it
        /// corresponding to identification rules.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <typeparam name="TIdentifier">The type of the identifier.</typeparam>
        /// <param name="instance">The instance to get identifier.</param>
        /// <returns>The instance identifier.</returns>
        /// <exception cref="PersistingRuntimeException">Error occured during get identifier.</exception>
        public virtual TIdentifier GetIdentifier<TInstance, TIdentifier>(TInstance instance)
            where TInstance : class
        {
            return BaseProvider.GetIdentifier<TInstance, TIdentifier>(instance);
        }

        /// <summary>
        /// Remember new identifier for instance registered in session.
        /// Call AcceptChanges method to actually apply new identifier to instance and
        /// register new identifier in session.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <typeparam name="TIdentifier">The type of the identifier.</typeparam>
        /// <param name="instance">The instance to update its identifier.</param>
        /// <param name="newIdentifier">The new identifier.</param>
        /// <exception cref="PersistingRuntimeException">Error occured during update identifier.</exception>
        /// <remarks>
        /// Identifier updating is required if identifiers are generated by database.
        /// Delayed identifier updating is required because if error occured due updating the instance
        /// or other instance in batch, identifier value generated by database becomes invalid.
        /// </remarks>
        public virtual void PutNewIdentifier<TInstance, TIdentifier>(TInstance instance, TIdentifier newIdentifier)
            where TInstance : class
        {
            BaseProvider.PutNewIdentifier<TInstance, TIdentifier>(instance, newIdentifier);
        }

        /// <summary>
        /// Gets entity with specified identifier. If entity registered in session it will be returned.
        /// Otherwise it will loaded from data storage. If there is no entity with specified identifier in data storage,
        /// method returns default value of <typeparamref name="TInstance"/> type.
        /// </summary>
        /// <typeparam name="TInstance"></typeparam>
        /// <typeparam name="TIdentifier"></typeparam>
        /// <param name="identifier">The identifier of entity to get from registry.</param>
        /// <returns>
        /// The entity with specified identifier or default value of <typeparamref name="TEntity"/> type of there is no
        /// entity with specified identifier exists in data storage.
        /// </returns>
        public virtual TInstance GetInstanceOrDefault<TInstance, TIdentifier>(TIdentifier identifier)
            where TInstance : class
        {
            return BaseProvider.GetInstanceOrDefault<TInstance, TIdentifier>(identifier);
        }

        /// <summary>
        /// Get object registered in session.
        /// If object is not found, the object will be constructed, registered in session, reloaded and
        /// returned.
        /// If object does not registered in session, it will be constructed
        /// using data extracted from default gateway.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <typeparam name="TIdentifier">The type of the identifier.</typeparam>
        /// <param name="identifier">The identifier of instance to get.</param>
        /// <returns>The instance with specified identifier.</returns>
        /// <exception cref="PersistingRuntimeException">Error occured during get object by identifier.</exception>
        public virtual TInstance GetInstance<TInstance, TIdentifier>(TIdentifier identifier)
            where TInstance : class
        {
            return BaseProvider.GetInstance<TInstance, TIdentifier>(identifier);
        }

        /// <summary>
        /// Get object registered in session.
        /// If object is not found, the object will be constructed, registered in session and reloaded
        /// using data passed as <paramref name="data"/> argument.
        /// If <paramref name="data"/> is null, the object will be constructed using data extracted from default gateway.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <typeparam name="TIdentifier">The type of the identifier.</typeparam>
        /// <param name="identifier">The identifier of instance to get.</param>
        /// <param name="data">The data used for object construction or null for using data from default gateway.</param>
        /// <returns>The instance with specified identifier.</returns>
        /// <exception cref="PersistingRuntimeException">Error occured during get object by identifier.</exception>
        public virtual TInstance GetInstance<TInstance, TIdentifier>(TIdentifier identifier, OverStore.Runtime.Storage.TypedRecord data)
            where TInstance : class
        {
            return BaseProvider.GetInstance<TInstance, TIdentifier>(identifier, data);
        }

        /// <summary>
        /// Return enumerator throught object sequence.
        /// </summary>
        /// <typeparam name="TItem">The type of the item.</typeparam>
        /// <typeparam name="TParameter">The type of the parameter for command.</typeparam>
        /// <param name="parameter">The parameter for command.</param>
        /// <returns>
        /// The sequence of object read from data storage.
        /// </returns>
        /// <exception cref="PersistingRuntimeException">Error occured while use object reader.</exception>
        /// <exception cref="ParameterTypeNotSupportedException">Parameter with specified type is not supported.</exception>
        public virtual IEnumerable<TItem> ExecuteObjectReader<TItem, TParameter>(TParameter parameter)
            where TItem : class
            where TParameter : class
        {
            return BaseProvider.ExecuteObjectReader<TItem, TParameter>(parameter);
        }

        /// <summary>
        /// Return enumerator throught object set.
        /// </summary>
        /// <typeparam name="TItem">The type of the item.</typeparam>
        /// <returns>
        /// The sequence of object read from data storage.
        /// </returns>
        /// <exception cref="PersistingRuntimeException">Error occured during creating object reader.</exception>
        public virtual IEnumerable<TItem> ExecuteObjectReader<TItem>()
            where TItem : class
        {
            return BaseProvider.ExecuteObjectReader<TItem>();
        }

        /// <summary>
        /// Saves changes made in every instace registered in data storage and invokes saving instance's aggregates.
        /// If instance was previously marked as deleted, method removes it from data storage.
        /// </summary>
        /// <exception cref="PersistingRuntimeException">Error occured while save object.</exception>
        /// <exception cref="VersionConflictException">Version conflict occured during save object.</exception>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same identifier already registered.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public virtual void SaveAll()
        {
            BaseProvider.SaveAll();
        }
    }

    /// <summary>
    /// Attach collection and instances before calling method of original provider.
    /// </summary>
    public sealed class AttachingPersistenceServiceProvider : CustomPersistenceServiceProvider
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="AttachingPersistenceServiceProvider"/> class.
        /// </summary>
        /// <param name="baseProvider">The base provider.</param>
        public AttachingPersistenceServiceProvider(IPersistenceServiceProvider baseProvider)
            : base(baseProvider)
        {
        }

        /// <summary>
        /// Get identifier for object registered in session.
        /// If object is not registered in session, it will be placed in session.
        /// If object does not have identifier yet, the identifier will be assigned to it
        /// corresponding to identification rules.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <typeparam name="TIdentifier">The type of the identifier.</typeparam>
        /// <param name="instance">The instance to get identifier.</param>
        /// <returns>The instance identifier.</returns>
        /// <exception cref="PersistingRuntimeException">Error occured during get identifier.</exception>
        public override TIdentifier GetIdentifier<TInstance, TIdentifier>(TInstance instance)
        {
            var result = BaseProvider.Attach<TInstance>(instance);
            return BaseProvider.GetIdentifier<TInstance, TIdentifier>(result);
        }

        /// <summary>
        /// Clears collection and fills it up with object read from parameterless reader for instance.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <exception cref="IdentityReferenceIntegrityException">Collection contains instance of object with same key already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error is occured during reload object data.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public override void ReloadCollection<TInstance>(IEnumerable<TInstance> collection)
        {
            var c = BaseProvider.AttachCollection<TInstance>(collection);
            BaseProvider.ReloadCollection<TInstance>(c);
        }

        /// <summary>
        /// Clears collection and fills it up with object read from parameterless reader for instance.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <typeparam name="TParameter">The type of the parameter.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <param name="parameter">The parameter used for get data for loading.</param>
        /// <exception cref="IdentityReferenceIntegrityException">Collection contains instance of object with same key already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error is occured during reload object data.</exception>
        /// <exception cref="ParameterTypeNotSupportedException">Parameters of specified type not supported.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public override void ReloadCollection<TInstance, TParameter>(IEnumerable<TInstance> collection, TParameter parameter)
        {
            var c = BaseProvider.AttachCollection<TInstance>(collection);
            BaseProvider.ReloadCollection<TInstance, TParameter>(collection, parameter);
        }

        /// <summary>
        /// 	<para>
        /// Saves changes made in each item of collection to data storage and invokes saving instance's aggregates.
        /// If instance was previously marked as deleted, method removes it from data storage.
        /// Also method saves instances were removed from collection, if collection type persister supports this feature.
        /// </para>
        /// 	<para>
        /// If collection supports deffer loading and is not loaded yet, method does not perform any
        /// actions, because changes can not be made without collection loading.
        /// </para>
        /// 	<para>
        /// Method does not accept collection changes, but types implements current method should consider possibility
        /// of calling this method few times with same instance, in case of cycle references, for example.
        /// Repeated calls should not cause repeated saving, so method should take steps to avoid repeated saving.
        /// For example, instance may internally maintain list of saved instances, checks this list on saving and
        /// clear this list in see cref="IRepository{TInstance}.AcceptChanges" method.
        /// </para>
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <exception cref="PersistingRuntimeException">Error occured while save object.</exception>
        /// <exception cref="VersionConflictException">Version conflict occured during save object.</exception>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same identifier already registered.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public override void SaveCollection<TInstance>(IEnumerable<TInstance> collection)
        {
            var c = BaseProvider.AttachCollection<TInstance>(collection);
            BaseProvider.SaveCollection<TInstance>(c);
        }

        /// <summary>
        /// Marks every item in collection as deleted. Method does not clears collection content.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <remarks>
        /// Method itself does not save any changes to storage. For actually delete <paramref name="instance"/>
        /// from storage, call <see cref="IRepository{TInstance}.Save"/> method.
        /// </remarks>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same identifier already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error occured while mark object as deleted.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public override void MarkCollectionAsDeleted<TInstance>(IEnumerable<TInstance> collection)
        {
            var c = BaseProvider.AttachCollection<TInstance>(collection);
            BaseProvider.MarkCollectionAsDeleted<TInstance>(c);
        }

        /// <summary>
        /// Marks every item in collection as needed to be physically saved to storage even if item data are not changed.
        /// This method may be called any times untill AcceptChanges method called, but instance will be saved only once.
        /// </summary>
        /// <typeparam name="TInstance">Type of instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <remarks>
        /// Use this method in couple with optimistic locking (version checking) to organize objects hierarchy into 'atom' group.
        /// Call this method on successful instance saving using other members of 'atom' object group, then save these objects.
        /// </remarks>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with the same identifier already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error occured while mark object as dirty.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public override void MarkCollectionAsDirty<TInstance>(IEnumerable<TInstance> collection)
        {
            var c = BaseProvider.AttachCollection<TInstance>(collection);
            BaseProvider.MarkCollectionAsDirty<TInstance>(c);
        }

        /// <summary>
        /// 	<para> Registers every item of collection in session , if it is not registered yet.
        /// After registration item become available by its identifier. </para>
        /// 	<para> If same instance already registered in repository, method does not perform any action.</para>
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <param name="collection">The collection.</param>
        /// <exception cref="IdentityReferenceIntegrityException">Different instance with same identifier already registered.</exception>
        /// <exception cref="PersistingRuntimeException">Error occured during add object.</exception>
        /// <exception cref="PersistenceOperationIsNotSupportedException">Operation is not supported because of configuration part responsible for this operation is missing or does not support specific functionallity.</exception>
        public override void AddCollection<TInstance>(IEnumerable<TInstance> collection)
        {
            var c = BaseProvider.AttachCollection<TInstance>(collection);
            BaseProvider.AddCollection<TInstance>(c);
        }
    }

    /// <summary>
    /// Provider that saves instance on trying to obtain their identifier.
    /// </summary>
    public sealed class SavingPersistenceServiceProvider : CustomPersistenceServiceProvider
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="SavingPersistenceServiceProvider"/> class.
        /// </summary>
        /// <param name="baseProvider">The base provider.</param>
        public SavingPersistenceServiceProvider(IPersistenceServiceProvider baseProvider)
            : base(baseProvider)
        {
        }

        /// <summary>
        /// Get identifier for object registered in session.
        /// If object is not registered in session, it will be placed in session.
        /// If object does not have identifier yet, the identifier will be assigned to it
        /// corresponding to identification rules.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <typeparam name="TIdentifier">The type of the identifier.</typeparam>
        /// <param name="instance">The instance to get identifier.</param>
        /// <returns>The instance identifier.</returns>
        /// <exception cref="PersistingRuntimeException">Error occured during get identifier.</exception>
        public override TIdentifier GetIdentifier<TInstance, TIdentifier>(TInstance instance)
        {
            BaseProvider.Save<TInstance>(instance);
            return BaseProvider.GetIdentifier<TInstance, TIdentifier>(instance);
        }
    }

    /// <summary>
    /// Provider that adds instance on trying to get their identifier.
    /// </summary>
    public sealed class AddingPersistenceServiceProvider : CustomPersistenceServiceProvider
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="AddingPersistenceServiceProvider"/> class.
        /// </summary>
        /// <param name="baseProvider">The base provider.</param>
        public AddingPersistenceServiceProvider(IPersistenceServiceProvider baseProvider)
            : base(baseProvider)
        {
        }

        /// <summary>
        /// Get identifier for object registered in session.
        /// If object is not registered in session, it will be placed in session.
        /// If object does not have identifier yet, the identifier will be assigned to it
        /// corresponding to identification rules.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance.</typeparam>
        /// <typeparam name="TIdentifier">The type of the identifier.</typeparam>
        /// <param name="instance">The instance to get identifier.</param>
        /// <returns>The instance identifier.</returns>
        /// <exception cref="PersistingRuntimeException">Error occured during get identifier.</exception>
        public override TIdentifier GetIdentifier<TInstance, TIdentifier>(TInstance instance)
        {
            BaseProvider.Add<TInstance>(instance);
            return BaseProvider.GetIdentifier<TInstance, TIdentifier>(instance);
        }
    }   
}
